@@ -1,4 +1,4 @@
-/* Copyright (c) 2006-2024 Jonas Fonseca <jonas.fonseca@gmail.com>
+/* Copyright (c) 2006-2025 Jonas Fonseca <jonas.fonseca@gmail.com>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -73,25 +73,23 @@ open_external_viewer(const char *argv[], const char *dir, bool silent, bool conf
 	bool ok;
 
 	if (echo) {
+		struct io io;
 		char buf[SIZEOF_STR] = "";
 
-		io_run_buf(argv, buf, sizeof(buf), dir, false);
-		if (*buf) {
+		ok = io_exec(&io, IO_RD, dir, NULL, argv, IO_RD_WITH_STDERR) && io_read_buf(&io, buf, sizeof(buf), true);
+		if (*buf)
 			report("%s", buf);
-			return true;
-		} else {
-			report("No output");
-			return false;
-		}
+
 	} else if (silent || is_script_executing()) {
 		ok = io_run_bg(argv, dir);
 
 	} else {
+		signal(SIGINT, SIG_IGN);
 		clear();
 		refresh();
 		endwin();                  /* restore original tty modes */
 		tcsetattr(opt_tty.fd, TCSAFLUSH, opt_tty.attr);
-		ok = io_run_fg(argv, dir);
+		ok = io_run_fg(argv, dir, opt_tty.fd);
 		if (confirm || !ok) {
 			if (!ok && *notice)
 				fprintf(stderr, "%s", notice);
@@ -104,6 +102,7 @@ open_external_viewer(const char *argv[], const char *dir, bool silent, bool conf
 		fseek(opt_tty.file, 0, SEEK_END);
 		tcsetattr(opt_tty.fd, TCSAFLUSH, opt_tty.attr);
 		set_terminal_modes();
+		signal(SIGINT, SIG_DFL);
 	}
 
 	if (watch_update(WATCH_EVENT_AFTER_COMMAND) && do_refresh) {
@@ -130,11 +129,9 @@ open_external_viewer(const char *argv[], const char *dir, bool silent, bool conf
 void
 open_editor(const char *file, unsigned int lineno)
 {
-	const char *editor_argv[SIZEOF_ARG + 3] = { "vi", file, NULL };
 	char editor_cmd[SIZEOF_STR];
-	char lineno_cmd[SIZEOF_STR];
+	const char *editor_argv[] = { "sh", "-c", editor_cmd, NULL };
 	const char *editor;
-	int argc = 0;
 
 	editor = getenv("TIG_EDITOR");
 	if (!editor)
@@ -148,15 +145,16 @@ open_editor(const char *file, unsigned int lineno)
 	if (!editor)
 		editor = "vi";
 
-	string_ncopy(editor_cmd, editor, strlen(editor));
-	if (!argv_from_string_no_quotes(editor_argv, &argc, editor_cmd)) {
-		report("Failed to read editor command");
-		return;
-	}
+	if (lineno && opt_editor_line_number)
+		string_format(editor_cmd, "%s +%u '%s'",
+			editor,
+			lineno,
+			file);
+	else
+		string_format(editor_cmd, "%s '%s'",
+			editor,
+			file);
 
-	if (lineno && opt_editor_line_number && string_format(lineno_cmd, "+%u", lineno))
-		editor_argv[argc++] = lineno_cmd;
-	editor_argv[argc] = file;
 	if (!open_external_viewer(editor_argv, repo.cdup, false, false, false, false, true, EDITOR_LINENO_MSG))
 		opt_editor_line_number = false;
 }
